import numpy as np
#from sklearn.model_selection import train_test_split
from tensorflow import keras
import csv
import os


# featurePath = '../data/features/pixels/'
# refusedPath = '../data/refuse/pixels/'

# for filename in os.listdir(featurePath):
#     if filename.endswith('.npy'):
#         dataArray = np.load(os.path.join(featurePath, filename))
#         # dataShape = np.shape(dataArray)
#         # dataCount = 1
#         # for dim in dataShape:
#         #     dataCount *= dim
#         # data_1D = np.reshape(dataArray, dataCount)
#         data_1D = dataArray.flatten()
#         dataAugmented = np.insert(data_1D, 0, 1)   # Insert a 1 at index 0 indicating this row as a feature
#         with open('pixel_data.csv','a') as csv_data:
#             csv_writer = csv.writer(csv_data)
#             csv_writer.writerow(dataAugmented)

# for filename in os.listdir(refusedPath):
#     if filename.endswith('.npy'):
#         dataArray = np.load(os.path.join(refusedPath, filename))
        
#         data_1D = dataArray.flatten()
#         dataAugmented = np.insert(data_1D, 0, 0)   # Insert a 0 at index 0 marking this row as refused 
#         with open('pixel_data.csv','a') as csv_data:
#             csv_writer = csv.writer(csv_data)
#             csv_writer.writerow(dataAugmented)

print('Data preparation completed.')



img_rows, img_cols = 20, 20
num_classes = 2

def prep_data(raw):
    y = raw[:, 0]
    out_y = keras.utils.to_categorical(y, num_classes)
    
    x = raw[:,1:]
    num_images = raw.shape[0]
    out_x = x.reshape(num_images, img_rows, img_cols, 1)
    #out_x = out_x / 255
    return out_x, out_y

fashion_file = "./pixel_data.csv"
fashion_data = np.loadtxt(fashion_file, skiprows=1, delimiter=',')
x, y = prep_data(fashion_data)

# Set up code checking
#from learntools.core import binder
#binder.bind(globals())
#from learntools.deep_learning.exercise_7 import *
print("Setup Complete")

#-----------------------------------------------------------------
from tensorflow import keras
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Flatten, Conv2D


fashion_model = Sequential()

fashion_model.add(Conv2D(12, kernel_size=(3,3), activation='relu', input_shape=(img_rows, img_cols, 1)))

fashion_model.add(Conv2D(20, kernel_size=(3,3), activation='relu'))
fashion_model.add(Conv2D(20, kernel_size=(3,3), activation='relu'))
fashion_model.add(Flatten())
fashion_model.add(Dense(100, activation='relu'))
fashion_model.add(Dense(num_classes, activation='softmax'))

fashion_model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])

fashion_model.fit(x, y, batch_size=100, epochs=4, validation_split=0.2)

#-----------------------------------------------------------------






import numpy as np


frame_1 = '/home/minter/workdir/Central_backup/Pictures/Test_images_for_estare_Ramberget_Dec_2020/DSC00281.JPG'

a_1 = [1270, 1953]
b_1 = [1289, 2087]

np.save('pivot.npy', (a_1, b_1))

frame_2 = '/home/minter/workdir/Central_backup/Pictures/Test_images_for_estare_Ramberget_Dec_2020/DSC00282.JPG'

a_2 = [1281, 1969]
b_2 = [1301, 2102]

np.save('target.npy', (a_2, b_2))




"""This is the entry point.

This script parses the input arguments, and depending on the selected
options, calls the relevant entry point function, i.e. scan() or 
assemble() with the parsed arguments. 
For more information on usage, run estare -h, or see the documentation.
"""
import argparse
from estare.src.modes import scan, assemble

       
parser = argparse.ArgumentParser(prog='estare', description='''Program for aligning and stacking 
astronomical photos.''', epilog='''estare is a Persian word for star.''')

sub_parsers = parser.add_subparsers()

# create the parser for the feature detection command
parse_feature = sub_parsers.add_parser('scan', help='''Scan an input image to find suitable features 
such as (a) bright star(s) that can be used for aligning multiple frames.''')

parse_feature.add_argument('--img1', action='store', type=str, required=True, \
                           help='Path and name of the reference image (base layer)')

parse_feature.add_argument('--img2', action='store', type=str, required=True, \
                           help='Path and name of the second image (top layer)')

parse_feature.add_argument('-i', '--info-only', action='store_true', \
                           help='''If specified, the properties of the 
                           input image will be printed to stdout, and the 
                           feature detection step will be skipped.''' )

parse_feature.add_argument('--save-pxs', '-x', action='store_true', help='''If set, the pixel values of the 
                                                                            feature(s) will be saved.''')
parse_feature.add_argument('--save-gif', '-g', action='store_true', help='''If set, the feature(s) will be 
                                                                            saved also in .png format.''')

parse_feature.set_defaults(func=scan)

parse_stack = sub_parsers.add_parser('stack', help='''This mode should normally be used after running 
the scan mode whereby alignment features are extracted. The stack mode uses the coordinates of two 
features, which are common in both frames, to align them. It then adjusts them and stacks them up. Two 
input images along with two identical features for each image are required.''')

parse_stack.add_argument('layer_1', help='Path and name of the reference image (base layer)')
parse_stack.add_argument('layer_2', help='Path and name of the second image (top layer)')
parse_stack.add_argument('feature_1', help='''Path and name of a .npy file containing the coordinates 
of the features in the reference layer''')
parse_stack.add_argument('feature_2', help='''Path and name of a .npy file containing the coordinates 
of the features in the second layer which will be lined up with the reference layer.''')

parse_stack.set_defaults(func=assemble)

args = parser.parse_args()


def main():
    args = parser.parse_args()
    args.func(args)   # call the default function

if __name__ == '__main__':
    main()



import unittest
import numpy as np
from skimage.io import imsave, imread
import os
from estare.src import align



class TestAlign(unittest.TestCase):
    

    @classmethod
    def setUpClass(self):
        from tests.reference_markers import reference_markers
        os.makedirs('./tests/test_results', exist_ok=True)
        targets, markers = reference_markers()
        np.save('./tests/test_images/pivot.npy', targets)
        np.save('./tests/test_images/target.npy', markers)


    @classmethod
    def tearDownClass(self):        

        for garbage_file in os.listdir('./tests/test_images'):
            os.remove(f'./tests/test_images/{garbage_file}')
            
        for garbage_file in os.listdir('./tests/test_results'):
            os.remove(f'./tests/test_results/{garbage_file}')            


        
    def test_offset(self):        
        
        del_x, del_y, theta, theta_rad = align.find_offset('./tests/test_images/pivot.npy',
                                                           './tests/test_images/target.npy')

        self.assertAlmostEqual(del_x, 26.74381464587602,
                               msg="Offset in the x-dir. is wrong.")
        
        self.assertAlmostEqual(del_y, -5.370829292101147,
                               msg="Offset in the y-dir. is wrong.")
        
        self.assertAlmostEqual(theta, 0.481598055901971,
                               msg="Rotation angle in degrees is wrong.")
        
        self.assertAlmostEqual(theta_rad, -0.008405471746693104,
                               msg="Rotation angle n radians is wrong.")


    def test_detrans(self):

        x_range = 400
        y_range = 300
        
        x_array = np.ones((x_range, y_range), dtype=int)
        y_array = np.ones((x_range, y_range), dtype=int)

        del_x = 1
        del_y = 1

        x_detrans, y_detrans = align.detrans(x_array, y_array, x_range,
                                             y_range, del_x, del_y)

        #TODO: check the entire arrays using numpy tests
        self.assertEqual(x_detrans[0, 0], -1)

        x_detrans, y_detrans = align.detrans_scalar(x_array, y_array, x_range,
                                             y_range, del_x, del_y)

        #TODO: check the entire arrays using numpy tests
        self.assertEqual(x_detrans[0, 0], -1)
        
        
    def test_derotate(self):
              
        del_x, del_y, theta, theta_rad = align.find_offset('./tests/test_images/pivot.npy',
                                                           './tests/test_images/target.npy')
        
        test_size=4
        
        x_array = np.zeros( (test_size, test_size), dtype=float )
        y_array = np.zeros( (test_size, test_size), dtype=float )
                
        for i in range(test_size):
            x_array[:,i] = [0, 1, 2, 3]
            y_array[i,:] = [0, 1, 2, 3]
        
        #TODO: rotate/derotate should not be applied to the anchor (origin)
        x_array_, y_array = align.derotate(x_array, y_array, theta, roundup=False)
        
        x_desired = np.array([[0, 0.00840537, 0.01681075, 0.02521612],
                     [0.99996467, 1.00837005, 1.01677542, 1.02518079],
                     [1.99992935, 2.00833472, 2.01674009, 2.02514547],
                     [2.99989402, 3.0082994, 3.01670477, 3.02511014]])
        
        y_desired = np.array( [[None, 0.99996467, 1.99992935, 2.99989402],
                               [-0.00840537, 0.9915593, 1.99152398, 2.99148865],
                               [-0.01681075, 0.98315393, 1.9831186, 2.98308328],
                               [-0.02521612, 0.97474856, 1.97471323, 2.9746779]] )
    
    
        np.testing.assert_almost_equal(x_array[1:, 1:], x_desired[1:, 1:])
        np.testing.assert_almost_equal(y_array[1:, 1:], y_desired[1:, 1:])


    def test_align(self):
        frame_0 = np.zeros( (270, 480), dtype=float )
       
        # pick 4 points as references
        position_1 = np.array( [200,100] ).T
        position_2 = np.array( [257,34] ).T 
        position_3 = np.array( [20,30] ).T
        position_4 = np.array( [123, 56] ).T

        # save two marker points for later stacking
        np.save('./tests/test_results/target.npy', [position_1, position_2])
        
        # mark the selected pixels in the original frame
        frame_0[position_1[0]-5:position_1[0]+5, position_1[1]-5:position_1[1]+5] = 1
        frame_0[position_2[0]-5:position_2[0]+5, position_2[1]-5:position_2[1]+5] = 1
        frame_0[position_3[0]-5:position_3[0]+5, position_3[1]-5:position_3[1]+5] = 1
        frame_0[position_4[0]-5:position_4[0]+5, position_4[1]-5:position_4[1]+5] = 1

        # save the base layer
        imsave('./tests/test_results/orig.jpeg', frame_0)
        
        # Now translate the rotated points
        transH = 15
        transW = 10

        pos1_tra = [position_1[0] + transH, position_1[1] + transW]
        pos2_tra = [position_2[0] + transH, position_2[1] + transW]
        pos3_tra = [position_3[0] + transH, position_3[1] + transW]
        pos4_tra = [position_4[0] + transH, position_4[1] + transW]

        np.testing.assert_equal(pos1_tra, [215, 110])
        np.testing.assert_equal(pos2_tra, [272,  44])
        np.testing.assert_equal(pos3_tra, [35, 40])
        np.testing.assert_equal(pos4_tra, [138,  66])

        # and now rotate them 20 deg.
        pos1_rot_tra = align.rotate(pos1_tra, 20)
        pos2_rot_tra = align.rotate(pos2_tra, 20)
        pos3_rot_tra = align.rotate(pos3_tra, 20)
        pos4_rot_tra = align.rotate(pos4_tra, 20)

        np.save('./tests/test_results/pivot.npy', [pos1_rot_tra, pos2_rot_tra])

        np.testing.assert_almost_equal(pos1_rot_tra, [164, 177])
        np.testing.assert_almost_equal(pos2_rot_tra, [241, 134])
        np.testing.assert_almost_equal(pos3_rot_tra, [19, 50])
        np.testing.assert_almost_equal(pos4_rot_tra, [107, 109])

        # Modified frame
        frame_1 = np.zeros_like(frame_0)
        frame_1[pos1_rot_tra[0]-5:pos1_rot_tra[0]+5,
                pos1_rot_tra[1]-5:pos1_rot_tra[1]+5] = 1
        
        frame_1[pos2_rot_tra[0]-5:pos2_rot_tra[0]+5,
                pos2_rot_tra[1]-5:pos2_rot_tra[1]+5] = 1
        
        frame_1[pos3_rot_tra[0]-5:pos3_rot_tra[0]+5,
                pos3_rot_tra[1]-5:pos3_rot_tra[1]+5] = 1
        
        frame_1[pos4_rot_tra[0]-5:pos4_rot_tra[0]+5,
                pos4_rot_tra[1]-5:pos4_rot_tra[1]+5] = 1

        # save the perturbed layer
        imsave('./tests/test_results/perturbed.jpeg', frame_1)
        
        # pick two of the rotated and translated points to back calculate the angle of rot.
        slope_bef = (position_2[1]-position_1[1]) / (position_2[0]-position_1[0])
        np.testing.assert_almost_equal((np.arctan(slope_bef))*180/np.pi, -49.18491612511842)

        slope_aft = (pos2_rot_tra[1]-pos1_rot_tra[1]) / (pos2_rot_tra[0]-pos1_rot_tra[0])
        np.testing.assert_almost_equal((np.arctan(slope_aft))*180/np.pi, -29.18080605249985)

        theta = (np.arctan(slope_aft))*180/np.pi - (np.arctan(slope_bef))*180/np.pi
        np.testing.assert_almost_equal(theta, 20.004110072618566)

        # derotate one of the references 
        pos1_derot = align.rotate(pos1_rot_tra, -theta)

        # compare the derotated one from its origin to calculate the relative translation  
        dxp = pos1_derot[0] - position_1[0]
        dyp = pos1_derot[1] - position_1[1]

        # Using the formulae, compute the absolute translation magnitude
        theta_rad = -theta * np.pi / 180   # radians
        del_y = ( dyp - dxp*np.tan(theta_rad) ) / \
                (np.sin(theta_rad)*np.tan(theta_rad) + np.cos(theta_rad))
        
        del_x = (dxp + del_y*np.sin(theta_rad)) / np.cos(theta_rad)

        np.testing.assert_almost_equal(del_x, 10.67414575051391)
        np.testing.assert_almost_equal(del_y, 14.52799409749281)

        # Now we can detranslate all four references
        pos1_detran = ([pos1_rot_tra[0] - del_x, pos1_rot_tra[1] - del_y] )
        pos2_detran = ([pos2_rot_tra[0] - del_x, pos2_rot_tra[1] - del_y] )
        pos3_detran = ([pos3_rot_tra[0] - del_x, pos3_rot_tra[1] - del_y] )
        pos4_detran = ([pos4_rot_tra[0] - del_x, pos4_rot_tra[1] - del_y] )

        # and now we can derotate them
        pos1_derot = align.rotate(pos1_detran, -theta)
        pos2_derot = align.rotate(pos2_detran, -theta)
        pos3_derot = align.rotate(pos3_detran, -theta)
        pos4_derot = align.rotate(pos4_detran, -theta)
       
        np.testing.assert_almost_equal(pos1_derot, [200, 100])
        np.testing.assert_almost_equal(pos2_derot, [257, 33])
        np.testing.assert_almost_equal(pos3_derot, [20,30])
        np.testing.assert_almost_equal(pos4_derot, [123, 56])

        # Reconstructed frame
        frame_3 = np.zeros_like(frame_0)

        frame_3[pos1_derot[0]-5:pos1_derot[0]+5, pos1_derot[1]-5:pos1_derot[1]+5] = 0.5
        frame_3[pos2_derot[0]-5:pos2_derot[0]+5, pos2_derot[1]-5:pos2_derot[1]+5] = 0.5
        frame_3[pos3_derot[0]-5:pos3_derot[0]+5, pos3_derot[1]-5:pos3_derot[1]+5] = 0.5
        frame_3[pos4_derot[0]-5:pos4_derot[0]+5, pos4_derot[1]-5:pos4_derot[1]+5] = 0.5

        stacked_image = align.align('./tests/test_results/orig.jpeg', './tests/test_results/perturbed.jpeg', './tests/test_results/target.npy', './tests/test_results/pivot.npy')
        #control = imread('../tests/stacked.jpeg')

        mis_align = np.linalg.norm(stacked_image - frame_0)
        #TODO: this test is problematic. Goodness of alignment changes with the algorithm
        self.assertLessEqual(np.abs(mis_align - 21.421162557285534), 1.0) 




import numpy as np



'''
We have:
- a function to calculate angle recieves a reference(default 0, 0) and a feature coordinate
- a function that takes an angle and a pair of coordinates and rotates the point (below)
- a function that takes [del_x, del_y] and a pair of coordinates and translates the point (below)

--------------->  Y
|*
| *
|  *
|   *
|
v

X

'''

# @profile
# def arc(position, origo=np.array([0., 0.]), radians=False):
#     '''
# Returns the angle between the input coordinate and the vertical axis Y
# '''
#     rho = np.sqrt( (position[0]-origo[0])**2 + (position[1]-origo[1])**2 )
#     angle = np.arccos( (position[0]-origo[0]) / rho )
    
#     if radians==False:
#         angle = angle*180.0/np.pi
        
#     return angle

def arc(position, origo=np.array([0., 0.]), radians=False):
    """Returns the angle between the input coordinate and the vertical 
       axis Y using the projection principle
    """

    #TODO: position[0,0] or generally position=origin should raise error
    # in fact division by rho = 0 should raise error
    position_norm = np.sqrt((position[0]-origo[0])**2 + (position[1]-origo[1])**2)
    # computing norm via np.linalg.norm(position) is much slower
    if radians==False:
        return position_norm, 180.0 * (np.arccos( (position[0]-origo[0]) / position_norm )) / np.pi
    else:
        return position_norm, np.arccos( (position[0]-origo[0]) / position_norm )

    
#@profile
def arc_vectorized(x_array, y_array, origo=np.array([0., 0.]), radians=False):
    """Returns the angle between the input coordinate and the vertical 
       axis Y using the projection principle
    """

    #TODO: position[0,0] or generally position=origin should raise error
    # in fact division by rho = 0 should raise error
    position_norm = np.sqrt((x_array-origo[0])**2 + (y_array-origo[1])**2)
    # computing norm via np.linalg.norm(position) is much slower
    if radians==False:
        return position_norm, ( (180.0/np.pi) * (np.arccos( (x_array-origo[0]) / position_norm )) )
    else:
        return position_norm, np.arccos( (x_array-origo[0]) / position_norm )

    
# apply rotation
#@profile
def rotate_vectorized(x_array, y_array, deflection, origo=np.array([0., 0.]), radians=False, discrete=True):
   """
   #TODO: position [0,0] cannot be rotated
   Change log:

   28/03/2021
   rho (which stands for 2-norm of the position vector) now comes directly from arc()
    
   02/04/2021
   rho and deflection are pre-stored in a look-up table, which means that they just need to
   be loaded

   11/09/2021 A vectorized version of arc() and rotate() is implemented
   """
   rho, theta = arc_vectorized(x_array, y_array, radians=radians)
   theta += deflection
   #theta = deflection + arc(position, radians=radians)

   # Use the compiled version of arc():
   # theta = deflection + calculate.get_angle(position, radians=radians)

   if radians==False:
       theta = theta*(np.pi/180.0)   

       # rho now comes directly from arc()
       #rho = np.sqrt((position[0]-origo[0])**2 + (position[1]-origo[1])**2)
    
   x = rho*np.cos(theta)
   y = rho*np.sin(theta)

   if discrete:
       rotated_x = np.rint(x)
       rotated_y = np.rint(y)     
       #        return  int(rotated_x), int(rotated_y)
       return  rotated_x, rotated_y
   else:
       return  x, y

    

def rotate(position, deflection, origo=np.array([0., 0.]), radians=False, discrete=True):
    """
    #TODO: position [0,0] cannot be rotated
    Change log:
    28/03/2021
    rho (which stands for 2-norm of the position vector) now comes directly from arc()
    
    02/04/2021
    rho and deflection are pre-stored in a look-up table, which means that they just need to
    be loaded
    """
    rho, theta = arc(position, radians=radians)
    theta += deflection
    #theta = deflection + arc(position, radians=radians)

    # Use the compiled version of arc():
    # theta = deflection + calculate.get_angle(position, radians=radians)

    if radians==False:
        theta = theta*(np.pi/180.0)   

    # rho now comes directly from arc()
    #rho = np.sqrt((position[0]-origo[0])**2 + (position[1]-origo[1])**2)
    
    x = rho*np.cos(theta)
    y = rho*np.sin(theta)

    if discrete:
        rotated_position = np.rint([x, y])     
        return np.array( [int(rotated_position[0]), int(rotated_position[1])] )
    else:
        return np.array( [x, y] )
    






"""This module should be imported by the entry-point script.

It implements the functions scan() and assemble(). The former 
is used for detecting features that are suitable for aligning 
two images. The latter proceeds to the alignment and stacking 
steps.
"""

from estare.src.align import align
from estare.src.init import examine, setup

from matplotlib import pyplot as plt
from skimage.io import imsave

import os
import numpy as np


def scan(args):
    """Finds features that are suitable for aligning two images.

    This function shows the input image in a graph first. It then 
    goes through every pixel of it and compares its brightness with 
    a given threshold (kapa) specified by the user. If the pixel's 
    brightness is higher than or equal to the threshold, a marker 
    will be shown on that location in the graph and the user can 
    accept it as a feature by pressing any key. The pixel can 
    alternatively be discarded by a left click.

    Args:
        args : user inputs parsed by argparse
    """
    
    # handle input arguments
    image1_path = args.img1
    image2_path = args.img2
    
    info_only = args.info_only   # just check image properties and skip feature detection
    save_pixs = args.save_pxs    # save also the pixel values
    save_gifs = args.save_gif    # save the features as PNG too
    
    if (info_only):   #TODO: this feature should have a separate argument subparser
        img_as_array, x_range, y_range = examine(image1_path, verbose=True, graphics=False)
        img_as_array, x_range, y_range = examine(image2_path, verbose=True, graphics=False)
        return 
    else:
        setup()   # setup the program directory structure
        img1_as_array, x_range, y_range = examine(image1_path, verbose=False, graphics=False)
        img2_as_array, x_range, y_range = examine(image2_path, verbose=False, graphics=False)
    # Replace with:
    # img_as_array = FloatImage(imagePath)
    # img_as_array.print_info()
    # img_as_array.save()
    # x_range, y_range = img_as_array.size()

    img1_gray = img1_as_array @ [0.2126, 0.7152, 0.0722]  # image in grayscale
    img2_gray = img2_as_array @ [0.2126, 0.7152, 0.0722]  
        
    # Replace with (makes FloatImage redundant):
    # imgGray = GrayImage(imagePath)
    # imgGray.print_info()
    # imgGray.save()
    # x_range, y_range = imgGray.size()

    fig2, (frame1,frame2) = plt.subplots(1, 2)
    image_features = frame1.imshow(img1_gray, cmap='gray')
    image_features = frame2.imshow(img2_gray, cmap='gray')
    selection_guide = '''Please first select two features from the left panel via left-click, and then pick their 
    counterparts (in the same order) from the right panel. Right-click to drop the latest selection. To interrupt 
    feature selection, use the middle mouse button.'''
    selection_warning='''Warning: zooming in can increase the accuracy of selection, but by doing this an 
    unwanted feature may be picked due to the mouse click action. In this case, right-click immediately after 
    zooming in to remove the unwanted selection.'''
    
    frame1.set_title(selection_guide)    
    frame2.set_title(selection_warning)
    
    plt.waitforbuttonpress(-1)
    coordinates = plt.ginput(n=4, timeout=-1, show_clicks=True)
    if len(coordinates) != 4:
        print('Feature detection incomplete. Please retry.')
    else:
        pivot_1 = [coordinates[0], coordinates[1]]
        pivot_2 = [coordinates[2], coordinates[3]]
        np.save('./estare_data/features/coordinates/pivot_1.npy', pivot_1)
        np.save('./estare_data/features/coordinates/pivot_2.npy', pivot_2)
        print('Feature detection completed.')



def assemble(args):
    """unpacks input arguments, then passes them to align() 

    This function prepares the input arguments to be passed to align() 
    for the algorithm to proceed to alignment and then stacking steps.

    Args:
        args : input arguments parsed by argparse
    """

    # TODO: design a stack-only mode that bypasses alignment
    
    # unpack the arguments
    img_1 = args.layer_1
    img_2 = args.layer_2
    pivot_1 = args.feature_1
    pivot_2 = args.feature_2

    # align and stack the two input frames
    stacked_frame = align(img_1, img_2, pivot_1, pivot_2, save=True)
    



import setuptools
#TODO Fix the Cython lib
from distutils import core
from distutils.extension import Extension
#from Cython.Distutils import build_ext



with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="estare", 
    version="1.0.0",
    author="Soheil Soltani",    
    description="Package for image alignment and stacking which can be used from the linux command-line",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/soheil-soltani/estare",
    packages=setuptools.find_packages(exclude="tests",),    
    classifiers=[
        "Programming Language :: Python :: 3",
	"License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX :: Linux",
    ],
    install_requires=[
        'numpy>=1.19.2',
        'matplotlib>=3.3.2',
        'scikit-image>=0.17.2',        
    ],
    python_requires='>=3.6',
    entry_points = {
        'console_scripts' : ['estare=estare.__main__:main']
    }
)

#core.setup(
#    cmdclass = {'build_ext' : build_ext},
#    ext_modules = [Extension("calculate", ["estare/src/arc.pyx"])]
#    )





import numpy as np
from estare.src.rotate import rotate_vectorized


def stack(img_1, img_2, x_array, y_array, x_range, y_range, deflection, discrete=True):
    """ theta was changed to deflection
    """
    
    ###angle_matrix = np.load('../aux/angle_table.npy')
    ###norm_matrix  = np.load('../aux/norm_matrix.npy')
    ###rho = norm_matrix[i, j]
    ###theta = angle_matrix[i, j]
    
    x_corrected, y_corrected = rotate_vectorized(x_array, y_array, -deflection, origo=np.array([0., 0.]), radians=False, discrete=discrete)
    
    for i in range(x_range):
        for j in range(y_range):
            # stack needs unittesting if the stacking algorithm is to be changed+
            #if x_array[i, j] > 0 and x_array[i, j] < x_range and y_array[i, j] > 0 and y_array[i, j] < y_range:
            #    img_1[x_array[i, j], y_array[i, j]] += img_2[i, j]   # Broadcasting to all three channels is implicit
            # Can this be vectorized too?    
            if (x_corrected[i,j] > 0) and (x_corrected[i,j] < x_range) and (y_corrected[i,j] > 0) and (y_corrected[i,j] < y_range):
                img_1[int(x_corrected[i,j]), int(y_corrected[i,j])] += img_2[i, j]   # Broadcasting to all three channels is implicit

    return img_1
                
            



from skimage import io
from skimage import img_as_float as imfloat
import numpy as np
from matplotlib import pyplot as plt

from estare.src.init import examine
from estare.src.rotate import rotate, rotate_vectorized
from estare.src.stack import stack

# +++++++++++++
# For profiling
import time


# Fine-tuning
# - pick features from scattered regions
# - characterize error: does it vary if one goes towards the edges or far from the origin?
# - compare light curves and make adjustments if the stacked feature deviates too much


def detrans(x_array, y_array, x_range, y_range, del_x, del_y):

    """Corrects the effect of translation by removing the extent by which each point 
    had been shifted, i.e. del_x, and del_y.

    Args:
    x_array
    y_array
    x_range
    y_range
    del_x
    del_y

    Returns:
    x_array
    y_array

    """
    
    for i in range(x_range):
        x_array[i, :] = i-del_x

    for j in range(y_range):
        y_array[:, j] = j-del_y
   
    
    return x_array, y_array



def detrans_scalar(x_array, y_array, x_range, y_range, del_x, del_y):
    
    for i in range(x_range):
        for j in range(y_range):
            x_array[i, j] = i-del_x
            y_array[i, j] = j-del_y
            
    
    return x_array, y_array



def find_offset(pivot_1, pivot_2):
    # two reference points for alignment
    coord_1 = np.load(pivot_1)
    a_1 = coord_1[0]
    b_1 = coord_1[1]
    
    # two target points that should match the reference points after alignment
    coord_2 = np.load(pivot_2)  
    a_2 = coord_2[0]
    b_2 = coord_2[1]

    # Use the two points to back calculate the angle of rotation
    slope_bef = ( b_1[1] - a_1[1] )/( b_1[0] - a_1[0] )
    
    slope_aft = ( b_2[1] - a_2[1] )/( b_2[0] - a_2[0] )

    deflection = ( ((np.arctan(slope_aft))*180/np.pi) -
              ( (np.arctan(slope_bef))*180/np.pi ) )

    a2_derot = rotate(a_2, -deflection, discrete=False)   # set discrete=False to get an exact value

    # compare the derotated one from its origin to calculate the relative translation magnitude 
    dxp = a2_derot[0] - a_1[0]
    dyp = a2_derot[1] - a_1[1]

    # Using the formulae, compute the absolute translation magnitude
    deflection_rad = -deflection * np.pi / 180   # radians

    del_y = ( dyp - dxp*np.tan(deflection_rad) ) / \
            (np.sin(deflection_rad)*np.tan(deflection_rad) + np.cos(deflection_rad))

    del_x = (dxp + del_y*np.sin(deflection_rad)) / np.cos(deflection_rad)


    return del_x, del_y, deflection, deflection_rad



def derotate(x_array, y_array, theta, roundup=False):
    """Removes the effect of rotation
    """

    x_range_1 = np.shape(x_array)[0]
    y_range_1 = np.shape(x_array)[1]
    
    for i in range(x_range_1):
        for j in range(y_range_1):
            uncorrected_coor = [x_array[i, j], y_array[i, j]]
            corrected_coor = rotate(uncorrected_coor, -theta, discrete=roundup)
            
            x_array[i, j] = corrected_coor[0]
            y_array[i, j] = corrected_coor[1]

    return x_array, y_array

                
def align(image_1, image_2, pivot_1, pivot_2, save=False):
    print('Working... this can take several minutes')
    #TODO: remove the save option. You always wanna save the result!
    img_1, x_range_1, y_range_1 = examine(image_1)  
    img_2, x_range_2, y_range_2 = examine(image_2)

    # if (x_range_1 != x_range_2) or (y_range_1 != y_range_2):
    #     #TODO: raise error and exit


    # calculate offsets (del_x, del_y) and rotation angle theta
    del_x, del_y, theta, theta_rad = find_offset(pivot_1, pivot_2)
    
    # Allocate x- and y-array for holding the new coordinates after offseting
    x_array = np.zeros((x_range_1, y_range_1, 1), dtype=int)
    y_array = np.zeros((x_range_1, y_range_1, 1), dtype=int)

    # For the two following time-consuming loops, we need unit tests to ensure the correctness after each optimization:
    # Now we can detranslate all points
    t_1 = time.time()

    x_array, y_array = detrans(x_array, y_array, x_range_1, y_range_1, del_x, del_y)

    t_2 = time.time()
    # and now we can derotate them using the vectorized version:
    stack(img_1, img_2, x_array, y_array, x_range_1, y_range_1, theta, discrete=True)
    t_3 = time.time()
    #TODO: configure profiling run separately from production
    #print('De-translation took %s sec.'%(t_2-t_1))
    #print('Stacking took %s sec.'%(t_3-t_2))    

    if save:
        io.imsave('./estare_data/stacked_images/estare_stacked_result.jpg', img_1)
        print('Finished. The result is saved as estare_stacked_result.jpg under ./estare_data/stacked_images/')
    
    return img_1
    




if __name__ == '__main__':

    x_range = 400; y_range = 600
    
    x_array = np.zeros((x_range, y_range, 1), dtype=float)
    y_array = np.zeros((x_range, y_range, 1), dtype=float)

    del_x = 0.1; del_y = 0.2

    
    t_1 = time.time()
    for t in range(10):
        x_array, y_array = detrans(x_array, y_array, x_range, y_range, del_x, del_y)
            
    t_2 = time.time()

    print('De-translation took %s sec.'%(t_2-t_1))

    
    t_1 = time.time()
    for t in range(10):
        x_array, y_array = detrans_scalar(x_array, y_array, x_range, y_range, del_x, del_y)
            
    t_2 = time.time()

    print('De-translation (scal. version) took %s sec.'%(t_2-t_1))


    
